@@ -13,6 +13,9 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/vfs.h>
+#include <sys/time.h>
+#include <sys/resource.h>
+#include <sys/wait.h>
 
 #include <sys/sendfile.h>
 
@@ -38,6 +41,7 @@
 #include "crtools.h"
 #include "cr_options.h"
 #include "servicefd.h"
+#include "string.h"
 #include "syscall.h"
 #include "ptrace.h"
 #include "util.h"
@@ -75,11 +79,12 @@
 #include "aio.h"
 #include "security.h"
 #include "lsm.h"
+#include "seccomp.h"
+#include "seize.h"
+#include "fault-injection.h"
 
 #include "asm/dump.h"
 
-#define NR_ATTEMPTS 5
-
 static char loc_buf[PAGE_SIZE];
 
 static void close_vma_file(struct vma_area *vma)
@@ -90,6 +95,8 @@ static void close_vma_file(struct vma_area *vma)
 		return;
 	if (vma->file_borrowed)
 		return;
+	if (vma_area_is(vma, VMA_AREA_AIORING))
+		return;
 
 	close(vma->vm_file_fd);
 }
@@ -233,7 +240,7 @@ static int dump_task_exe_link(pid_t pid, MmEntry *mm)
 	struct fd_parms params;
 	int fd, ret = 0;
 
-	fd = open_proc(pid, "exe");
+	fd = open_proc_path(pid, "exe");
 	if (fd < 0)
 		return -1;
 
@@ -256,7 +263,7 @@ static int dump_task_fs(pid_t pid, struct parasite_dump_misc *misc, struct cr_im
 	fe.has_umask = true;
 	fe.umask = misc->umask;
 
-	fd = open_proc(pid, "cwd");
+	fd = open_proc_path(pid, "cwd");
 	if (fd < 0)
 		return -1;
 
@@ -271,7 +278,7 @@ static int dump_task_fs(pid_t pid, struct parasite_dump_misc *misc, struct cr_im
 
 	close(fd);
 
-	fd = open_proc(pid, "root");
+	fd = open_proc_path(pid, "root");
 	if (fd < 0)
 		return -1;
 
@@ -339,7 +346,8 @@ static int dump_filemap(pid_t pid, struct vma_area *vma_area,
 	if (vma_area->aufs_rpath) {
 		struct fd_link aufs_link;
 
-		strcpy(aufs_link.name, vma_area->aufs_rpath);
+		strlcpy(aufs_link.name, vma_area->aufs_rpath,
+				sizeof(aufs_link.name));
 		aufs_link.len = strlen(aufs_link.name);
 		p.link = &aufs_link;
 	}
@@ -534,7 +542,7 @@ static int get_task_personality(pid_t pid, u32 *personality)
 {
 	int fd, ret = -1;
 
-	pr_info("Obtaining personality ... ");
+	pr_info("Obtaining personality ... \n");
 
 	fd = open_proc(pid, "personality");
 	if (fd < 0)
@@ -663,6 +671,7 @@ static int dump_task_core_all(struct pstree_item *item,
 	CoreEntry *core = item->core[0];
 	pid_t pid = item->pid.real;
 	int ret = -1;
+	struct proc_status_creds *creds;
 
 	pr_info("\n");
 	pr_info("Dumping core (pid: %d)\n", pid);
@@ -672,7 +681,19 @@ static int dump_task_core_all(struct pstree_item *item,
 	if (ret < 0)
 		goto err;
 
-	strncpy((char *)core->tc->comm, stat->comm, TASK_COMM_LEN);
+	creds = dmpi(item)->pi_creds;
+	if (creds->seccomp_mode != SECCOMP_MODE_DISABLED) {
+		pr_info("got seccomp mode %d for %d\n", creds->seccomp_mode, item->pid.virt);
+		core->tc->has_seccomp_mode = true;
+		core->tc->seccomp_mode = creds->seccomp_mode;
+
+		if (creds->seccomp_mode == SECCOMP_MODE_FILTER) {
+			core->tc->has_seccomp_filter = true;
+			core->tc->seccomp_filter = creds->last_filter;
+		}
+	}
+
+	strlcpy((char *)core->tc->comm, stat->comm, TASK_COMM_LEN);
 	core->tc->flags = stat->flags;
 	core->tc->task_state = item->state;
 	core->tc->exit_code = 0;
@@ -701,316 +722,6 @@ err:
 	return ret;
 }
 
-static int parse_children(pid_t pid, pid_t **_c, int *_n)
-{
-	pid_t *ch = NULL;
-	int nr = 0;
-	DIR *dir;
-	struct dirent *de;
-
-	dir = opendir_proc(pid, "task");
-	if (dir == NULL)
-		return -1;
-
-	while ((de = readdir(dir))) {
-		int fd, len;
-		char *pos;
-
-		if (dir_dots(de))
-			continue;
-
-		fd = open_proc(pid, "task/%s/children", de->d_name);
-		if (fd < 0)
-			goto err;
-
-		len = read(fd, loc_buf, sizeof(loc_buf) - 1);
-		close(fd);
-		if (len < 0)
-			goto err;
-
-		loc_buf[len] = '\0';
-		pos = loc_buf;
-		while (1) {
-			pid_t val, *tmp;
-
-			val = strtol(pos, &pos, 0);
-			if (!val) {
-				BUG_ON(*pos != '\0');
-				break;
-			}
-
-			tmp = xrealloc(ch, (nr + 1) * sizeof(pid_t));
-			if (!tmp)
-				goto err;
-
-			ch = tmp;
-			ch[nr] = val;
-			nr++;
-			pos++; /* space goes after each pid */
-		}
-	}
-
-	*_c = ch;
-	*_n = nr;
-
-	closedir(dir);
-	return 0;
-err:
-	closedir(dir);
-	xfree(ch);
-	return -1;
-}
-
-static inline bool child_collected(struct pstree_item *i, pid_t pid)
-{
-	struct pstree_item *c;
-
-	list_for_each_entry(c, &i->children, sibling)
-		if (c->pid.real == pid)
-			return true;
-
-	return false;
-}
-
-static int collect_task(struct pstree_item *item);
-static int collect_children(struct pstree_item *item)
-{
-	pid_t *ch;
-	int ret, i, nr_children, nr_inprogress;
-
-	ret = parse_children(item->pid.real, &ch, &nr_children);
-	if (ret < 0)
-		return ret;
-
-	nr_inprogress = 0;
-	for (i = 0; i < nr_children; i++) {
-		struct pstree_item *c;
-		pid_t pid = ch[i];
-
-		/* Is it already frozen? */
-		if (child_collected(item, pid))
-			continue;
-
-		nr_inprogress++;
-
-		pr_info("Seized task %d, state %d\n", pid, ret);
-
-		c = alloc_pstree_item();
-		if (c == NULL) {
-			ret = -1;
-			goto free;
-		}
-
-		ret = seize_task(pid, item->pid.real);
-		if (ret < 0) {
-			/*
-			 * Here is a race window between parse_children() and seize(),
-			 * so the task could die for these time.
-			 * Don't worry, will try again on the next attempt. The number
-			 * of attempts is restricted, so it will exit if something
-			 * really wrong.
-			 */
-			ret = 0;
-			xfree(c);
-			continue;
-		}
-
-		c->pid.real = pid;
-		c->parent = item;
-		c->state = ret;
-		list_add_tail(&c->sibling, &item->children);
-
-		/* Here is a recursive call (Depth-first search) */
-		ret = collect_task(c);
-		if (ret < 0)
-			goto free;
-	}
-free:
-	xfree(ch);
-	return ret < 0 ? ret : nr_inprogress;
-}
-
-static void unseize_task_and_threads(const struct pstree_item *item, int st)
-{
-	int i;
-
-	if (item->state == TASK_DEAD)
-		return;
-
-	/*
-	 * The st is the state we want to switch tasks into,
-	 * the item->state is the state task was in when we seized one.
-	 */
-
-	unseize_task(item->pid.real, item->state, st);
-
-	for (i = 1; i < item->nr_threads; i++)
-		ptrace(PTRACE_DETACH, item->threads[i].real, NULL, NULL);
-}
-
-static void pstree_switch_state(struct pstree_item *root_item, int st)
-{
-	struct pstree_item *item = root_item;
-
-	pr_info("Unfreezing tasks into %d\n", st);
-	for_each_pstree_item(item)
-		unseize_task_and_threads(item, st);
-}
-
-static pid_t item_ppid(const struct pstree_item *item)
-{
-	item = item->parent;
-	return item ? item->pid.real : -1;
-}
-
-static inline bool thread_collected(struct pstree_item *i, pid_t tid)
-{
-	int t;
-
-	if (i->pid.real == tid) /* thread leader is collected as task */
-		return true;
-
-	for (t = 0; t < i->nr_threads; t++)
-		if (tid == i->threads[t].real)
-			return true;
-
-	return false;
-}
-
-static int collect_threads(struct pstree_item *item)
-{
-	struct pid *threads = NULL;
-	int nr_threads = 0, i = 0, ret, nr_inprogress, nr_stopped = 0;
-
-	ret = parse_threads(item->pid.real, &threads, &nr_threads);
-	if (ret < 0)
-		goto err;
-
-	if ((item->state == TASK_DEAD) && (nr_threads > 1)) {
-		pr_err("Zombies with threads are not supported\n");
-		goto err;
-	}
-
-	/* The number of threads can't be less than allready frozen */
-	item->threads = xrealloc(item->threads, nr_threads * sizeof(struct pid));
-	if (item->threads == NULL)
-		return -1;
-
-	if (item->nr_threads == 0) {
-		item->threads[0].real = item->pid.real;
-		item->nr_threads = 1;
-	}
-
-	nr_inprogress = 0;
-	for (i = 0; i < nr_threads; i++) {
-		pid_t pid = threads[i].real;
-
-		if (thread_collected(item, pid))
-			continue;
-
-		nr_inprogress++;
-
-		pr_info("\tSeizing %d's %d thread\n",
-				item->pid.real, pid);
-
-		ret = seize_task(pid, item_ppid(item));
-		if (ret < 0) {
-			/*
-			 * Here is a race window between parse_threads() and seize(),
-			 * so the task could die for these time.
-			 * Don't worry, will try again on the next attempt. The number
-			 * of attempts is restricted, so it will exit if something
-			 * really wrong.
-			 */
-			continue;
-		}
-
-		BUG_ON(item->nr_threads + 1 > nr_threads);
-		item->threads[item->nr_threads].real = pid;
-		item->nr_threads++;
-
-		if (ret == TASK_DEAD) {
-			pr_err("Zombie thread not supported\n");
-			goto err;
-		}
-
-		if (ret == TASK_STOPPED) {
-			nr_stopped++;
-		}
-	}
-
-	if (nr_stopped && nr_stopped != nr_inprogress) {
-		pr_err("Individually stopped threads not supported\n");
-		goto err;
-	}
-
-	xfree(threads);
-	return nr_inprogress;
-
-err:
-	xfree(threads);
-	return -1;
-}
-
-static int collect_loop(struct pstree_item *item,
-		int (*collect)(struct pstree_item *))
-{
-	int attempts = NR_ATTEMPTS, nr_inprogress = 1;
-
-	/*
-	 * While we scan the proc and seize the children/threads
-	 * new ones can appear (with clone(CLONE_PARENT) or with
-	 * pthread_create). Thus, after one go, we need to repeat
-	 * the scan-and-freeze again collecting new arrivals. As
-	 * new guys may appear again we do NR_ATTEMPTS passes and
-	 * fail to seize the item if new tasks/threads still
-	 * appear.
-	 */
-
-	while (nr_inprogress > 0 && attempts) {
-		attempts--;
-		nr_inprogress = collect(item);
-	}
-
-	/*
-	 * We may fail to collect items or run out of attempts.
-	 * In the former case nr_inprogress will be negative, in
-	 * the latter -- positive. Thus it's enough just to check
-	 * for "no more new stuff" and say "we're OK" if so.
-	 */
-
-	return (nr_inprogress == 0) ? 0 : -1;
-}
-
-static int collect_task(struct pstree_item *item)
-{
-	int ret;
-
-	ret = collect_loop(item, collect_threads);
-	if (ret < 0)
-		goto err_close;
-
-	/* Depth-first search (DFS) is used for traversing a process tree. */
-	ret = collect_loop(item, collect_children);
-	if (ret < 0)
-		goto err_close;
-
-	if ((item->state == TASK_DEAD) && !list_empty(&item->children)) {
-		pr_err("Zombie with children?! O_o Run, run, run!\n");
-		goto err_close;
-	}
-
-	if (pstree_alloc_cores(item))
-		goto err_close;
-
-	pr_info("Collected %d in %d state\n", item->pid.real, item->state);
-	return 0;
-
-err_close:
-	close_pid_proc();
-	return -1;
-}
-
 static int collect_pstree_ids_predump(void)
 {
 	struct pstree_item *item;
@@ -1052,36 +763,6 @@ int collect_pstree_ids(void)
 	return 0;
 }
 
-static int collect_pstree(pid_t pid)
-{
-	int ret;
-
-	timing_start(TIME_FREEZING);
-
-	root_item = alloc_pstree_item();
-	if (root_item == NULL)
-		return -1;
-
-	root_item->pid.real = pid;
-	ret = seize_task(pid, -1);
-	if (ret < 0)
-		goto err;
-	pr_info("Seized task %d, state %d\n", pid, ret);
-	root_item->state = ret;
-
-	ret = collect_task(root_item);
-	if (ret < 0)
-		goto err;
-
-	timing_stop(TIME_FREEZING);
-	timing_start(TIME_FROZEN);
-
-	return 0;
-err:
-	pstree_switch_state(root_item, TASK_ALIVE);
-	return -1;
-}
-
 static int collect_file_locks(void)
 {
 	return parse_file_locks();
@@ -1129,7 +810,7 @@ static int dump_one_zombie(const struct pstree_item *item,
 	if (!core)
 		return -1;
 
-	strncpy((char *)core->tc->comm, pps->comm, TASK_COMM_LEN);
+	strlcpy((char *)core->tc->comm, pps->comm, TASK_COMM_LEN);
 	core->tc->task_state = TASK_DEAD;
 	core->tc->exit_code = pps->exit_code;
 
@@ -1522,6 +1203,11 @@ static int dump_one_task(struct pstree_item *item)
 		goto err;
 	}
 
+	if (fault_injected(FI_DUMP_EARLY)) {
+		pr_info("fault: CRIU sudden detach\n");
+		BUG();
+	}
+
 	if (root_ns_mask & CLONE_NEWPID && root_item == item) {
 		int pfd;
 
@@ -1771,12 +1457,18 @@ int cr_dump_tasks(pid_t pid)
 {
 	struct pstree_item *item;
 	int post_dump_ret = 0;
+	int pre_dump_ret = 0;
 	int ret = -1;
 
 	pr_info("========================================\n");
 	pr_info("Dumping processes (pid: %d)\n", pid);
 	pr_info("========================================\n");
 
+	pre_dump_ret = run_scripts(ACT_PRE_DUMP);
+	if (pre_dump_ret != 0) {
+		pr_err("Pre dump script failed with %d!\n", post_dump_ret);
+		goto err;
+	}
 	if (init_stats(DUMP_STATS))
 		goto err;
 
@@ -1834,6 +1526,9 @@ int cr_dump_tasks(pid_t pid)
 	if (!glob_imgset)
 		goto err;
 
+	if (collect_seccomp_filters() < 0)
+		goto err;
+
 	for_each_pstree_item(item) {
 		if (dump_one_task(item))
 			goto err;
@@ -1875,7 +1570,6 @@ int cr_dump_tasks(pid_t pid)
 	if (ret)
 		goto err;
 
-	fd_id_show_tree();
 err:
 	if (disconnect_from_page_server())
 		ret = -1;
