@@ -3,10 +3,8 @@
 
 #include "protobuf/rpc.pb-c.h"
 
-#define CR_DEFAULT_SERVICE_ADDRESS "/var/run/criu_service.socket"
-#define MAX_MSG_SIZE 1024
-
 extern int cr_service(bool deamon_mode);
+int cr_service_work(int sk);
 
 extern int send_criu_dump_resp(int socket_fd, bool success, bool restored);
 
