@@ -6,6 +6,11 @@
  */
 
 #define CRTOOLS_IMAGES_V1	1
+/*
+ * v1.1 has common magic in the head of each image file,
+ * except for inventory
+ */
+#define CRTOOLS_IMAGES_V1_1	2
 
 /*
  * Raw images are images in which data is stored in some
@@ -14,6 +19,14 @@
 
 #define RAW_IMAGE_MAGIC		0x0
 
+/*
+ * Images have the IMG_COMMON_MAGIC in the head. Service files
+ * such as stats and irmap-cache have the IMG_SERVICE_MAGIC.
+ */
+
+#define IMG_COMMON_MAGIC	0x54564319 /* Sarov (a.k.a. Arzamas-16) */
+#define IMG_SERVICE_MAGIC	0x55105940 /* Zlatoust */
+
 /*
  * The magic-s below correspond to coordinates
  * of various Russian towns in the NNNNEEEE form.
@@ -21,7 +34,6 @@
 
 #define INVENTORY_MAGIC		0x58313116 /* Veliky Novgorod */
 #define PSTREE_MAGIC		0x50273030 /* Kyiv */
-#define STATS_MAGIC		0x57093306 /* Ostashkov */
 #define FDINFO_MAGIC		0x56213732 /* Dmitrov */
 #define PAGEMAP_MAGIC		0x56084025 /* Vladimir */
 #define SHMEM_PAGEMAP_MAGIC	PAGEMAP_MAGIC
@@ -61,6 +73,7 @@
 #define INOTIFY_WD_MAGIC	0x54562009 /* Svetlogorsk (Rauschen) */
 #define MNTS_MAGIC		0x55563928 /* Petushki */
 #define NETDEV_MAGIC		0x57373951 /* Yaroslavl */
+#define NETNS_MAGIC		0x55933752 /* Dolgoprudny */
 #define TTY_FILES_MAGIC		0x59433025 /* Pushkin */
 #define TTY_INFO_MAGIC		0x59453036 /* Kolpino */
 #define FILE_LOCKS_MAGIC	0x54323616 /* Kaluga */
@@ -76,16 +89,24 @@
 #define TIMERFD_MAGIC		0x50493712 /* Korocha */
 #define CPUINFO_MAGIC		0x61404013 /* Nyandoma */
 #define USERNS_MAGIC		0x55474906 /* Kazan */
+#define SECCOMP_MAGIC		0x64413049 /* Kostomuksha */
 
 #define IFADDR_MAGIC		RAW_IMAGE_MAGIC
 #define ROUTE_MAGIC		RAW_IMAGE_MAGIC
+#define ROUTE6_MAGIC		RAW_IMAGE_MAGIC
+#define RULE_MAGIC		RAW_IMAGE_MAGIC
 #define TMPFS_IMG_MAGIC		RAW_IMAGE_MAGIC
 #define TMPFS_DEV_MAGIC		RAW_IMAGE_MAGIC
 #define IPTABLES_MAGIC		RAW_IMAGE_MAGIC
+#define IP6TABLES_MAGIC		RAW_IMAGE_MAGIC
 
 #define PAGES_OLD_MAGIC		PAGEMAP_MAGIC
 #define SHM_PAGES_OLD_MAGIC	PAGEMAP_MAGIC
 
+/*
+ * These are special files, not exactly images
+ */
+#define STATS_MAGIC		0x57093306 /* Ostashkov */
 #define IRMAP_CACHE_MAGIC	0x57004059 /* Ivanovo */
 
 #endif /* __CR_MAGIC_H__ */
