@@ -19,7 +19,7 @@
 #include "timerfd.h"
 #include "shmem.h"
 #include "sigframe.h"
-#include "vdso.h"
+#include "parasite-vdso.h"
 
 #include <time.h>
 
@@ -90,7 +90,8 @@ struct thread_restore_args {
 	tls_t				tls;
 
 	siginfo_t			*siginfo;
-	unsigned int			siginfo_nr;
+	unsigned int			siginfo_n;
+
 	int				pdeath_sig;
 } __aligned(64);
 
@@ -103,16 +104,46 @@ struct task_restore_args {
 
 	/* threads restoration */
 	int				nr_threads;		/* number of threads */
-	int				nr_zombies;
 	thread_restore_fcall_t		clone_restore_fn;	/* helper address for clone() call */
 	struct thread_restore_args	*thread_args;		/* array of thread arguments */
 	struct task_entries		*task_entries;
 	void				*rst_mem;
 	unsigned long			rst_mem_size;
-	VmaEntry			*tgt_vmas;
+
+	/* Below arrays get remapped from RM_PRIVATE in sigreturn_restore */
+	VmaEntry			*vmas;
+	unsigned int			vmas_n;
+
+	struct restore_posix_timer	*posix_timers;
+	unsigned int			posix_timers_n;
+
+	struct restore_timerfd		*timerfd;
+	unsigned int			timerfd_n;
+
 	siginfo_t			*siginfo;
-	unsigned int			siginfo_nr;
-	unsigned int			nr_vmas;
+	unsigned int			siginfo_n;
+
+	struct rst_tcp_sock		*tcp_socks;
+	unsigned int			tcp_socks_n;
+
+	struct rst_aio_ring		*rings;
+	unsigned int			rings_n;
+
+	struct rlimit			*rlims;
+	unsigned int			rlims_n;
+
+	pid_t				*helpers /* the TASK_HELPERS to wait on at the end of restore */;
+	unsigned int			helpers_n;
+
+	pid_t				*zombies;
+	unsigned int			zombies_n;
+
+	struct sock_fprog		*seccomp_filters;
+	unsigned int			seccomp_filters_n;
+
+	/* * * * * * * * * * * * * * * * * * * * */
+
+	unsigned long			task_size;
 	unsigned long			premmapped_addr;
 	unsigned long			premmapped_len;
 	rt_sigaction_t			sigchld_act;
@@ -122,12 +153,6 @@ struct task_restore_args {
 
 	struct itimerval		itimers[3];
 
-	int				timer_n;
-	struct restore_posix_timer	*posix_timers;
-
-	int				timerfd_n;
-	struct restore_timerfd		*timerfd;
-
 	CredsEntry			creds;
 	u32				cap_inh[CR_CAP_SIZE];
 	u32				cap_prm[CR_CAP_SIZE];
@@ -140,23 +165,14 @@ struct task_restore_args {
 	u32				mm_saved_auxv_size;
 	char				comm[TASK_COMM_LEN];
 
-	int				nr_rlim;
-	struct rlimit			*rlims;
+	/*
+	 * proc_fd is a handle to /proc that the restorer blob can use to open
+	 * files there, because some of them can't be opened before the
+	 * restorer blob is called.
+	 */
+	int				proc_fd;
 
-	struct rst_tcp_sock		*tcp_socks;
-	int				tcp_socks_nr;
-
-	struct rst_aio_ring		*rings;
-	int				nr_rings;
-
-	int				fd_last_pid; /* sys.ns_last_pid for threads rst */
-
-	pid_t				*helpers /* the TASK_HELPERS to wait on at the end of restore */;
-	int				n_helpers;
-
-	int				proc_attr_current;
-	char				*lsm_profile;
-	int				lsm_profile_len;
+	int				seccomp_mode;
 
 #ifdef CONFIG_VDSO
 	unsigned long			vdso_rt_size;
