@@ -20,7 +20,7 @@
 #define xzalloc(size)		__xalloc(calloc, size, 1, size)
 #define xrealloc(p, size)	__xalloc(realloc, size, p, size)
 
-#define xfree(p)		do { if (p) free(p); } while (0)
+#define xfree(p)		free(p)
 
 #define xrealloc_safe(pptr, size)					\
 	({								\
