@@ -20,16 +20,19 @@ static char buf[512];
  * ANy brave soul to write it using xtables-devel?
  */
 
-static const char *nf_conn_cmd = "%s -t filter %s INPUT --protocol tcp "
+static const char *nf_conn_cmd = "%s -t filter %s %s --protocol tcp "
 	"--source %s --sport %d --destination %s --dport %d -j DROP";
 
 static char iptable_cmd_ipv4[] = "iptables";
 static char iptable_cmd_ipv6[] = "ip6tables";
 
-static int nf_connection_switch_raw(int family, u32 *src_addr, u16 src_port, u32 *dst_addr, u16 dst_port, int lock)
+static int nf_connection_switch_raw(int family, u32 *src_addr, u16 src_port,
+						u32 *dst_addr, u16 dst_port,
+						bool input, bool lock)
 {
 	char sip[INET_ADDR_LEN], dip[INET_ADDR_LEN];
 	char *cmd;
+	char *argv[4] = { "sh", "-c", buf, NULL };
 	int ret;
 
 	switch (family) {
@@ -50,11 +53,18 @@ static int nf_connection_switch_raw(int family, u32 *src_addr, u16 src_port, u32
 		return -1;
 	}
 
-	snprintf(buf, sizeof(buf), nf_conn_cmd, cmd, lock ? "-A" : "-D",
+	snprintf(buf, sizeof(buf), nf_conn_cmd, cmd,
+			lock ? "-A" : "-D",
+			input ? "INPUT" : "OUTPUT",
 			dip, (int)dst_port, sip, (int)src_port);
 
 	pr_debug("\tRunning iptables [%s]\n", buf);
-	ret = system(buf);
+
+	/*
+	 * cr_system is used here, because it blocks SIGCHLD before waiting
+	 * a child and the child can't be waited from SIGCHLD handler.
+	 */
+	ret = cr_system(-1, -1, -1, "sh", argv, 0);
 	if (ret < 0 || !WIFEXITED(ret) || WEXITSTATUS(ret)) {
 		pr_perror("Iptables configuration failed");
 		return -1;
@@ -65,34 +75,34 @@ static int nf_connection_switch_raw(int family, u32 *src_addr, u16 src_port, u32
 	return 0;
 }
 
-static int nf_connection_switch(struct inet_sk_desc *sk, int lock)
+static int nf_connection_switch(struct inet_sk_desc *sk, bool lock)
 {
 	int ret = 0;
 
 	ret = nf_connection_switch_raw(sk->sd.family,
 			sk->src_addr, sk->src_port,
-			sk->dst_addr, sk->dst_port, lock);
+			sk->dst_addr, sk->dst_port, true, lock);
 	if (ret)
 		return -1;
 
 	ret = nf_connection_switch_raw(sk->sd.family,
 			sk->dst_addr, sk->dst_port,
-			sk->src_addr, sk->src_port, lock);
+			sk->src_addr, sk->src_port, false, lock);
 	if (ret) /* rollback */
 		nf_connection_switch_raw(sk->sd.family,
 			sk->src_addr, sk->src_port,
-			sk->dst_addr, sk->dst_port, !lock);
+			sk->dst_addr, sk->dst_port, true, !lock);
 	return ret;
 }
 
 int nf_lock_connection(struct inet_sk_desc *sk)
 {
-	return nf_connection_switch(sk, 1);
+	return nf_connection_switch(sk, true);
 }
 
 int nf_unlock_connection(struct inet_sk_desc *sk)
 {
-	return nf_connection_switch(sk, 0);
+	return nf_connection_switch(sk, false);
 }
 
 int nf_unlock_connection_info(struct inet_sk_info *si)
@@ -101,10 +111,10 @@ int nf_unlock_connection_info(struct inet_sk_info *si)
 
 	ret |= nf_connection_switch_raw(si->ie->family,
 			si->ie->src_addr, si->ie->src_port,
-			si->ie->dst_addr, si->ie->dst_port, 0);
+			si->ie->dst_addr, si->ie->dst_port, true, false);
 	ret |= nf_connection_switch_raw(si->ie->family,
 			si->ie->dst_addr, si->ie->dst_port,
-			si->ie->src_addr, si->ie->src_port, 0);
+			si->ie->src_addr, si->ie->src_port, false, false);
 	/*
 	 * rollback nothing in case of any error,
 	 * because nobody checks errors of this function
