@@ -63,29 +63,24 @@ int ns_child(void *_arg)
 	}
 
 	if (stat(MPTS_ROOT"/dev/mntns2/test", &st)) {
-		err("Can't stat /dev/share-1/test.share/test.share");
+		pr_perror("Can't stat /dev/share-1/test.share/test.share");
 		return 1;
 	}
 
 	return 0;
 }
 
-static int test_fn(int argc, char **argv)
+int main(int argc, char **argv)
 {
 	FILE *f;
-	int fd, tmpfs_fd;
+	int fd, tmpfs_fd, have_bfmtm = 0;
 	unsigned fs_cnt, fs_cnt_last = 0;
 	struct ns_exec_args args;
 	mode_t old_mask;
 	pid_t pid = -1;
 
-	if (!getenv("ZDTM_REEXEC")) {
-		setenv("ZDTM_REEXEC", "1", 0);
-		return execv(argv[0], argv);
-	} else
-		test_init(argc, argv);
+	test_init(argc, argv);
 
-	close(0); /* /dev/null */
 again:
 	fs_cnt = 0;
 	f = fopen("/proc/self/mountinfo", "r");
@@ -95,7 +90,7 @@ again:
 	}
 
 	if (mount(NULL, "/", NULL, MS_REC|MS_PRIVATE, NULL)) {
-		err("Can't remount / with MS_PRIVATE");
+		pr_perror("Can't remount / with MS_PRIVATE");
 		return -1;
 	}
 
@@ -151,7 +146,7 @@ done:
 	}
 	tmpfs_fd = open(MPTS_ROOT"/dev/test", O_WRONLY | O_CREAT);
 	if (write(tmpfs_fd, "hello", 5) <= 0) {
-		err("write() failed");
+		pr_perror("write() failed");
 		return 1;
 	}
 
@@ -159,18 +154,18 @@ done:
 	mkdir(MPTS_ROOT"/dev/overmount", 0600);
 	fd = open(MPTS_ROOT"/dev/overmount/test.over", O_WRONLY | O_CREAT);
 	if (fd == -1) {
-		err("Unable to open "MPTS_ROOT"/dev/overmount\n");
+		pr_perror("Unable to open "MPTS_ROOT"/dev/overmount");
 		return -1;
 	}
 	close(fd);
 	if (mount("none", MPTS_ROOT"/dev/overmount", "tmpfs", 0, "") < 0) {
-		err("Can't mount "MPTS_ROOT"/dev/overmount\n");
+		pr_perror("Can't mount "MPTS_ROOT"/dev/overmount");
 		return 1;
 	}
 
 	mkdir(MPTS_ROOT"/dev/non-root", 0600);
 	if (mount(MPTS_ROOT"/dev/non-root", MPTS_ROOT"/module", NULL, MS_BIND, NULL) < 0) {
-		err("Can't bind-mount %s -> %s", MPTS_ROOT"/dev/tdir", MPTS_ROOT"/module");
+		pr_perror("Can't bind-mount %s -> %s", MPTS_ROOT"/dev/tdir", MPTS_ROOT"/module");
 	}
 	mkdir(MPTS_ROOT"/dev/non-root/test", 0600);
 
@@ -225,14 +220,14 @@ done:
 	}
 
 	mkdir(MPTS_ROOT"/dev/share-1/test.mnt.share", 0600);
-	if (mount("none", MPTS_ROOT"/dev/share-1/test.mnt.share", "tmpfs", 0, "") < 0) {
+	if (mount("none", MPTS_ROOT"/dev/share-1/test.mnt.share", "tmpfs", 0, "size=1G") < 0) {
 		fail("Can't mount tmpfs");
 		return 1;
 	}
 
 	mkdir(MPTS_ROOT"/dev/share-1/test.mnt.share/test.share", 0600);
 	if (umount(MPTS_ROOT"/dev/slave2/test.mnt.share")) {
-		err("Can't umount "MPTS_ROOT"/dev/slave2/test.mnt.share: %m");
+		pr_perror("Can't umount "MPTS_ROOT"/dev/slave2/test.mnt.share");
 		return 1;
 	}
 
@@ -245,14 +240,14 @@ done:
 
 	fd = open(MPTS_ROOT"/dev/bmfile", O_CREAT | O_WRONLY);
 	if (fd < 0) {
-		err("Can't create " MPTS_ROOT "/dev/share-1/bmfile");
+		pr_perror("Can't create " MPTS_ROOT "/dev/share-1/bmfile");
 		return 1;
 	}
 	close(fd);
 
 	fd = open(MPTS_ROOT"/dev/bmfile-mount", O_CREAT | O_WRONLY);
 	if (fd < 0) {
-		err("Can't create " MPTS_ROOT "/dev/share-1/bmfile");
+		pr_perror("Can't create " MPTS_ROOT "/dev/share-1/bmfile");
 		return 1;
 	}
 	close(fd);
@@ -268,10 +263,8 @@ done:
 	}
 
 	if (mount("none", MPTS_ROOT"/kernel/sys/fs/binfmt_misc",
-					"binfmt_misc", 0, "") < 0) {
-		fail("Can't mount binfmt_misc");
-		return 1;
-	}
+					"binfmt_misc", 0, "") == 0)
+		have_bfmtm = 1;
 
 	unlink("/dev/null");
 	/*
@@ -291,7 +284,7 @@ done:
 	if (getenv("ZDTM_NOSUBNS") == NULL) {
 		pid = clone(ns_child, args.stack_ptr, CLONE_NEWNS | SIGCHLD, &args);
 		if (pid < 0) {
-			err("Unable to fork child");
+			pr_perror("Unable to fork child");
 			return 1;
 		}
 	}
@@ -305,8 +298,13 @@ done:
 		return 1;
 	}
 
+	if (have_bfmtm && access(MPTS_ROOT"/kernel/sys/fs/binfmt_misc/register", F_OK)) {
+		fail("No binfmt_misc after restore");
+		return 1;
+	}
+
 	if (umount(MPTS_ROOT"/dev/overmount") == -1) {
-		err("Can't umount "MPTS_ROOT"/dev/overmount\n");
+		pr_perror("Can't umount "MPTS_ROOT"/dev/overmount");
 		return -1;
 	}
 	if (access(MPTS_ROOT"/dev/overmount/test.over", F_OK)) {
@@ -317,11 +315,11 @@ done:
 	{
 		struct stat st1, st2;
 		if (stat(MPTS_ROOT"/dev/share-1/test.mnt.share/test.share", &st1)) {
-			err("Can't stat /dev/share-1/test.share/test.share");
+			pr_perror("Can't stat /dev/share-1/test.share/test.share");
 			return 1;
 		}
 		if (stat(MPTS_ROOT"/dev/share-2/test.mnt.share/test.share", &st2)) {
-			err("Can't stat /dev/share-2/test.mnt.share/test.share");
+			pr_perror("Can't stat /dev/share-2/test.mnt.share/test.share");
 			return 1;
 		}
 		if (st1.st_ino != st2.st_ino) {
@@ -329,7 +327,7 @@ done:
 			return 1;
 		}
 		if (stat(MPTS_ROOT"/dev/slave/test.mnt.share/test.share", &st2)) {
-			err("Can't stat /dev/slave/test.mnt.share/test.share");
+			pr_perror("Can't stat /dev/slave/test.mnt.share/test.share");
 			return 1;
 		}
 		if (st1.st_ino != st2.st_ino) {
@@ -337,15 +335,15 @@ done:
 			return 1;
 		}
 		if (stat(MPTS_ROOT"/dev/share-1/test.mnt.slave/test.slave", &st1) != -1 || errno != ENOENT) {
-			err("/dev/share-1/test.mnt.slave/test.slave exists");
+			pr_perror("/dev/share-1/test.mnt.slave/test.slave exists");
 			return 1;
 		}
 		if (stat(MPTS_ROOT"/dev/slave/test.mnt.slave/test.slave", &st2)) {
-			err("Can't stat /dev/slave/test.mnt.slave/test.slave");
+			pr_perror("Can't stat /dev/slave/test.mnt.slave/test.slave");
 			return 1;
 		}
 		if (stat(MPTS_ROOT"/dev/non-root/test", &st1)) {
-			err("Can't stat /dev/non-root/test");
+			pr_perror("Can't stat /dev/non-root/test");
 			return 1;
 		}
 	}
@@ -361,13 +359,3 @@ done:
 	pass();
 	return 0;
 }
-
-#define CLONE_NEWNS     0x00020000
-
-int main(int argc, char **argv)
-{
-	if (getenv("ZDTM_REEXEC"))
-		return test_fn(argc, argv);
-	test_init_ns(argc, argv, CLONE_NEWNS, test_fn);
-	return 0;
-}
